import { isEmpty, omitBy, omit, mapValues, intersection, pull, findIndex, get, values } from 'lodash';
import { Http } from './http';
import * as storage from './storage'; 
import { changes } from './utils';

const _config = {
  base: '/',
  ext: '.json',
  storage: {
    prefix: 'snap:',
  },
};

// TODO: split some functions to utils
// TODO: keep local list as a single list

export class Snapmongo {
  static config = ({
    base,
    ext,
    storage: { prefix } = {},
  }) => {
    Object.assign(_config, {
      base,
      ext,
    });
    if (prefix) {
      Object.assign(_config.storage, { prefix });
    }
  }

  constructor (collection , conf = {
  }) {
    storage.setPrefix(_config.storage.prefix);
    Object.assign(this, {
      schema: {
        fields: [],
        detailFields: [],
      },
      name: '',
      isPending: false,
      isError: false,
      isReady: false,
      pendingResolves: [],
      pendingRejects: [],
      catalog: null,
      list: {},
      details: {},
      http: {},
      catalogName: 'index',
      config: {
        ext: _config.ext,
        base: _config.base,
      },
      httpOptions: {
        json: true,
        timestampsField: 'meta.timestamps',
        updatedField: 'meta.updated',
      }
    });
    this.name = collection;
    Object.assign(this.config, conf);
    this.http = new Http(this.config);
    this._requestCatalog()
      .catch(err => {
        // console.error(err);
        throw err;
      });
    storage.get('overrides:lists')
      .then(res => {
        if (Array.isArray(res)) {
          const [first] = res;
          storage.set('overrides:lists', null);
          storage.set(`overrides:list`, first);
        }
      })
      .catch();
  }

  _getFilename(name) {
    return this.config.base + '/' + this.name + '/' + name + this.config.ext;
  }
  _getLatest(name) {
    const timestamps = get(this.catalog, this.httpOptions.timestampsField) || {}; 
    return timestamps[name];
  }
  _addList(list) {
    Object.keys(list).forEach(id => {
      this.list[id] = { ...list[id], _id: id };
    });
  }
  _requestList(listName) {
    return this.http.get(this._getFilename(listName), {
      ...this.httpOptions,
      latest: this._getLatest(listName),
    })
      .then(data => {
        if (!data.list || !data.meta) {
          throw new Error('missing list and/or meta in the json');
        }
        this._addList(data.list);
        if (data.meta && data.meta.nextList) {
          return this._requestList(data.meta.nextList);
        }
      });
  }
  async _determineConflicts(remote) {
    // check the same data and remove local updates
    const [list] = await Promise.all([
      storage.get('overrides:list').catch(() => ({})),
      storage.get('overrides:details').catch(() => ({})),
    ]);
    const savedIds = Object.keys(list);
    const catalogIds = Object.keys(remote);
    const conflictIds = intersection(savedIds, catalogIds);
    // console.log('------conflictIds', conflictIds);
    let save;
    conflictIds.forEach(key => {
      const diff = omit(changes(remote[key], list[key]), ['_id']);
      if (isEmpty(diff)) {
        save = true;
        // console.log('identical!!!!', remote[key], list[key]);
        // TODO: check details also
        delete list[key];
      } else {
        // console.log('diff:', diff);
      }
    });
    if (save) {
      return storage.set('overrides:list', list);
    }
  }

  _requestCatalog() {
    if (this.isPending) {
      return new Promise((res, rej) => {
        this.pendingResolves.push(res);
        this.pendingRejects.push(rej);
      });
    }
    this.isPending = true;
    return this.http.get(this._getFilename(this.catalogName), {
      ...this.httpOptions,
      latest: true,
    })
      .then(async data => {
        if (!data.list || !data.meta) {
          throw new Error('missing list and/or meta in the json');
        }
        this.catalog = data;
        await this._determineConflicts(data.list);
        this._addList(data.list);
        Object.assign(this.details, data.details);
        const { meta } = data;
        if (meta.fields) {
          Object.assign(this.schema.fields, meta.fields);
        }
        if (meta.detailFields) {
          Object.assign(this.schema.detailFields, meta.detailFields);
        }
        let listPromise = Promise.resolve();
        if (this.catalog.meta.nextList) {
          listPromise = this._requestList(this.catalog.meta.nextList);
        }
        return listPromise.then(() => {
          this.hasDetails = !!this.catalog.meta.detailFields;
          this.isReady = true;
          this.isError = false;
          this.isPending = false;
          this.pendingResolves.forEach(res => res());
          this.pendingResolves = [];
          this.pendingRejects = [];
        });
      })
      .catch(err => {
        this.isError = true;
        this.isPending = false;
        this.pendingRejects.forEach(rej => rej(err));
        this.pendingResolves = [];
        this.pendingRejects = [];
        throw err;
      });
  }

  _getBrief(id, options = {}) {
    const brief = this.list[id];
    const localPromise = (
      options.remoteOnly
        ? Promise.resolve([ {} ])
        : storage.get('overrides:list').catch(() => ({}))
    );
    return localPromise.then(localList => {
      const localBrief = localList[id];
      if (!brief && !localBrief) {
        return null;
      }
      if (brief && this.hasDetails) {
        const detail = this.details[id];
        if (!detail && brief.hasDetail) {
          Object.assign(brief, { _missing: true });
        } else {
          Object.assign(brief, { _missing: false });
        }
      }
      return Object.assign({}, brief, localBrief);
    });
  }

  _getData(id, brief, options = {}) {
    if (!brief.hasDetail) {
      return Promise.resolve(Object.assign(brief, { _id: id }));
    }
    const localPromise2 = (
      options.remoteOnly
        ? Promise.resolve({})
        : storage.get('overrides:details').catch(() => ({}))
    );
    return localPromise2.then(localDetails => {
      const localDetail = localDetails[id];
      const detail = this.details[id];
      return Object.assign({}, brief, detail, localDetail, { _id: id });
    });
  }

  _getList({ localList, options = {} } = {}) {
    const remote = this.list;
    const localPromise = options.remoteOnly
      ? Promise.resolve({})
      : localList
        ? Promise.resolve(localList)
        : storage.get('overrides:list').catch(() => ({}));
    return localPromise.then(local => {
      const merged = omitBy(Object.assign({}, remote, local), it => it._deleted);
      // console.log('merged:', merged, values(merged).map(i=>i._id));
      const ret = values(merged);
      ret.reverse();
      ret.total = this.catalog.meta.total;
      return ret;
    });
  }

  find(q, options) {
    return this._requestCatalog()
      .then(() => {
        return this._getList({ options });
      });
  }

  findOne(q, options) {
    const id = get(q, '_id') || q;
    let catalogPromise = Promise.resolve();
    if (!this.isReady) {
      catalogPromise = this._requestCatalog();
    }
    return catalogPromise.then(async () => {
      const brief = await this._getBrief(id, options);
      if (!brief) {
        return null;
      }
      if (!brief._missing) {
        return this._getData(id, brief, options);
      }
      return this._requestData(brief)
        .then(() => {
          return this._getData(id, brief, options);
        });
    });
  }

  _requestData(brief) {
    return this.http.get(this._getFilename(brief._detail), {
      ...this.httpOptions,
      latest: this._getLatest(brief._detail),
    })
      .then(data => {
        Object.assign(this.details, data.details);
      });
  }

  _addArray(target, src) {
    const _src = [].concat(src);
    pull(_src, target);
    _src.forEach(it => target.push(it));
  }
  _getNewId(localList) {
    return this._getList({ localList })
      .then(list => {
        const max = list[0]._id;
        return `${parseInt(max, 10) + 1}`;
      });
  }
  _saveOverrides(list, details){
    return Promise.all([
      storage.set('overrides:list', list),
      storage.set('overrides:details', details),
    ]);
  }
  _updateSchema(fields, detailFields) {
    const { schema } = this;
    this._addArray(schema.fields, fields);
    this._addArray(schema.detailFields, detailFields);
  }
  _splitDocument(doc) {
    const details = doc._details ? doc._details : {};
    const detailFields = Object.keys(details);
    delete doc._details;
    const fields = Object.keys(doc);
    this._addArray(detailFields, intersection(fields, this.schema.detailFields));
    pull(fields, ...detailFields);
    return [ doc, details, fields, detailFields ];
  }
  insert(obj) {
    return Promise.all([
      storage.get('overrides:list').catch(() => ({})),
      storage.get('overrides:details').catch(() => ({})),
    ])
      .then(([ list, _details ]) => {
        return this._getNewId(list)
          .then(id => {
            const [
              brief,
              details,
              fields,
              detailFields
            ] = this._splitDocument(Object.assign({}, obj, { _id: id, _local: 'new' }));
            this._updateSchema(fields, detailFields);
            _details = {
              ..._details,
              [id]: details,
            };
            list[id] = brief;
            return this._saveOverrides(list, _details);
          });
      });
  }
  update(q, obj) {
    const id = `${get(q, '_id') || q}`;
    return Promise.all([
      storage.get('overrides:list').catch(() => ({})),
      storage.get('overrides:details').catch(() => ({})),
    ])
      .then(([ list, _details ]) => {
        // console.log(list, _details);
        const [
          brief,
          details,
          fields,
          detailFields
        ] = this._splitDocument(Object.assign({}, obj, { _id: id, _local: 'update' }));
        this._updateSchema(fields, detailFields);
        Object.assign(_details, { [id]: this.details[id] }, { [id]: details });
        const remoteList = this.list;
        Object.assign(list, remoteList && { [id]: remoteList[id] }, { [id]: brief });
        // console.log('after:', list, _details);
        return this._saveOverrides(list, _details);
      });
  }
  dumpUpdated() {
    return Promise.all([
      this._getList(),
      storage.get('overrides:list').catch(() => ({})),
      storage.get('overrides:details').catch(() => ({})),
    ])
      .then(([ orgList, localList, localDetails ]) => {
        if (Object.keys(localList).length === 0) {
          return {};
        }
        const meta = Object.assign({}, this.catalog.meta, {
          total: orgList.length,
          fields: this.schema.fields,
          detailFields: this.schema.detailFields,
        });
        const list = Object.assign(
          {},
          this.catalog.list,
          mapValues(localList, item => omit(item, '_local'))
        );
        const details = Object.assign({}, this.catalog.details, localDetails);
        return {
          filename: 'index.json',
          data: {
            meta,
            list,
            details,
          },
        };
      });
  }
  delete(q) {
    const id = `${get(q, '_id') || q}`;
    return Promise.all([
      storage.get('overrides:list').catch(() => ({})),
      storage.get('overrides:details').catch(() => ({})),
    ])
      .then(([list]) => {
        if (!list[id]) {
          list[id] = { _id: id };
        }
        list[id]._deleted = true;
        return storage.set('overrides:list', list);
      });
  }
}
export default Snapmongo;
