import chai, { expect } from 'chai';
import spies from 'chai-spies';
import { find, defer, forOwn, findIndex } from 'lodash';
import { Snapmongo } from '../src/snapmongo';
import storage from '../src/storage'; 
import { Http } from '../src/http';

chai.use(spies);

Snapmongo.config({
  base: '',
  ext: '.json',
});

describe('Snap Mongo', () => {
  let users;
  let spyOnFetch;
  before(done => {
    spyOnFetch = chai.spy.on(window, 'fetch');
    done();
  });
  describe('new object', () => {
    it('should have name and find', (done) => {
      users = new Snapmongo('users');
      expect(users.name).to.equal('users');
      expect(users.find).to.exist;
      expect(users.findOne).to.exist;
      users.find().then(() => done());
    });
    it('should throw error when there is no such collection', done => {
      const nocollection = new Snapmongo('nocollection');
      expect(nocollection.name).to.equal('nocollection');
      expect(nocollection.find).to.exist;
      expect(nocollection.findOne).to.exist;
      nocollection.find()
      .then(() => done('there was no exception'))
      .catch(err => {
        done();
      });
    });
  });

  describe('new Snapmongo', () => {
    it('should modify lists into list in the local storage', (done) => {
      localStorage.setItem('snap:overrides:lists', '[{}]');
      users = new Snapmongo('blank');
      users.find()
        .then(json => {
          const oldLists = localStorage.getItem('snap:overrides:lists');
          const newList = localStorage.getItem('snap:overrides:list');
          expect(oldLists).to.equal(null);
          expect(newList).to.equal('{}');
          done();
        })
        .catch(err => done(err));
    });
  });

  describe('find', () => {
    it('should return promise', () => {
      const promise = users.find();
      expect(promise.then).to.exist;
      expect(promise.catch).to.exist;
    });

    it('should be resolved with json', (done) => {
      users.find()
        .then(json => {
          expect(json).to.exist;
          expect(Array.isArray(json)).to.equal(true);
          expect(json.total).to.exist;
          done();
        })
        .catch(err => done(err));
    });

    it('should use timestamp in the catalog file', () => {
      // console.log(spyOnFetch.__spy.calls);
      const calls = spyOnFetch.__spy.calls;
      const [firstUrl] = calls[0];
      expect(firstUrl.startsWith('/users/index.json?')).to.be.true;
    });
  });

  describe('dumpUpdated', () => {
    it('should return empty object if there is no update', (done) => {
      const main2 = new Snapmongo('main');
      main2.find().then(() => {
        return main2.dumpUpdated().then(dump => {
          expect(Object.keys(dump).length).to.equal(0);
          done();
        });
      })
        .catch(err => done(err));
    });
  });

  describe('findOne', () => {
    it('should return promise', () => {
      const promise = users.findOne();
      expect(promise.then).to.exist;
      expect(promise.catch).to.exist;
    });

    it('should be resolved with json', (done) => {
      users.findOne(5)
        .then(json => {
          expect(json).to.satisfy(obj => obj === null || obj);
          expect(typeof json).to.equal('object');
          expect((json && json._id) || true).to.exist;
          done();
        })
        .catch(err => {
          console.error(err);
          done(err)
        });
    });
    it('should show merged data', (done) => {
      done();
    });
    it('should show only when the saved option is true', (done) => {
      done();
    });
  });
  describe('blog main retrieving', () => {
    let main;
    let list;
    before(() => {
      main = new Snapmongo('main');
    });
    beforeEach(function (done) {
      main.find().then(res => {
        list = res;
        done();
      })
      .catch(err => done(err));
    });
    it('should have propper name', () => {
      expect(main.name).to.equal('main');
    });
    it('should be able to find four items', () => {
      expect(list.length).to.equal(4);
    });
    it('should be able to find the oldest item', (done) => {
      const brief = list.pop();
      main.findOne(brief._id)
      .then(item => {
        expect(item).to.exist;
        done();
      })
      .catch(err => done(err));
    });
    it('should be able to find the latest item', (done) => {
      const brief = list.shift();
      main.findOne(brief._id)
      .then(item => {
        expect(item).to.exist;
        done();
      })
      .catch(err => done(err));
    });
  });

  describe('update', () => {
    let main;
    let list;
    before(() => {
      main = new Snapmongo('main');
    });
    beforeEach(function (done) {
      main.find().then(res => {
        list = res;
        done();
      })
      .catch(err => done(err));
    });
    it('should update fields', (done) => {
      main.update(4, { title: 'updated item' })
        .then(() => {
          return main.findOne(4)
            .then(item => {
              expect(item.title).to.equal('updated item');
              done();
            });
        })
        .catch(err => done(err));
    });
    it('should save the document in the local storage', (done) => {
      const list = JSON.parse(localStorage.getItem('snap:overrides:list'));
      // console.log(list);
      const item = list['4'];
      expect(item.title).to.equal('updated item');
      done();
    });
  });

  describe('insert', () => {
    let main;
    let list;
    before(() => {
      main = new Snapmongo('main');
    });
    beforeEach(function (done) {
      main.find().then(res => {
        list = res;
        done();
      })
      .catch(err => done(err));
    });
    it('should update fields in the meta if new key exists', (done) => {
      main.insert({ newField: 'test' })
        .then(() => {
          expect(main.schema.fields.indexOf('newField') >= 0).to.be.true;
          done();
        })
        .catch(err => done(err));
    });
    it('should update fields in the meta if new key is in _details', (done) => {
      main.insert({ _details: { newDetail: 'test' } })
        .then(() => {
          expect(main.schema.detailFields.indexOf('newDetail') >= 0).to.be.true;
          done();
        })
        .catch(err => done(err));
    });
    it('should split the document into list and detail according to schema', (done) => {
      main.find()
        .then(list => {
          list.forEach(item => {
            if (item.newDetail) {
              done('detail found in list');
            }
          });
          done();
        });
    });
    it('should create a document with _id in type string', (done) => {
      main.find()
        .then(list => {
          list.forEach(item => {
            expect(typeof item._id).to.equal('string');
          });
          done();
        })
        .catch(err => done(err));
    });
    it('should create a document with _local', (done) => {
      main.find()
        .then(list => {
          const idx = findIndex(list, it => it.newField);
          expect(idx).to.be.above(0);
          expect(list[idx]._local).to.exist;
          done();
        })
        .catch(err => done(err));
    });
    it('should save the document in the local storage', (done) => {
      const list = JSON.parse(localStorage.getItem('snap:overrides:list'));
      const details = JSON.parse(localStorage.getItem('snap:overrides:details'));
      let foundList = false;
      forOwn(list, item => {
        if (item.newField) {
          foundList = true;
        }
      });
      expect(foundList).to.be.true;
      let foundDetail = false;
      forOwn(details, item => {
        if (item.newDetail) {
          foundDetail = true;
        }
      });
      expect(foundDetail).to.be.true;
      done();
    });
    it('let find include inserted data', (done) => {
      main.find()
        .then(list => {
          const idx = findIndex(list, it => it.newField);
          expect(idx >= 0).to.be.true;
          done();
        })
        .catch(err => done(err));
    });
    it('let find show only remote date when the remoteOnly option is true', done => {
      main.find({}, { remoteOnly: true })
        .then(list => {
          // console.log(list);
          const idx = findIndex(list, it => it.newField);
          expect(idx < 0).to.be.true;
          done();
        })
        .catch(err => done(err));
    });
  });

  describe('dumpUpdated', () => {
    let main;
    let list;
    let dump;
    before(() => {
      main = new Snapmongo('main');
    });
    before(function (done) {
      main.insert({ newField: 'test' })
        .then(() => done())
        .catch(err => done(err));
    });
    beforeEach(function (done) {
      main.find().then(res => {
        list = res;
        main.dumpUpdated()
          .then(res => {
            dump = res;
            done();
          });
      })
      .catch(err => done(err));
    });
    it('should return object with filename and data', (done) => {
      expect(dump.filename).to.exist;
      expect(dump.data).to.exist;
      done();
    });
    it('should not return _local in any items', (done) => {
      const { meta, list, details } = dump;
      forOwn(list, item => {
        expect(item._local).to.be.not.exist;
      });
      done();
    });
    it('should not show updated items when the remoteOnly option is true after dumping', (done) => {
      main.find({}, { remoteOnly: true })
        .then(list => {
          const idx = findIndex(list, it => it.newField);
          expect(idx < 0).to.be.true;
          done();
        })
        .catch(err => done(err));
    });
  });
  describe('delete', () => {
    let main;
    let list;
    before(() => {
      main = new Snapmongo('main');
    });
    beforeEach(function (done) {
      main.find().then(res => {
        list = res;
        done();
      })
      .catch(err => done(err));
    });
    it('should add _deleted to the item and ignore that while find or findOne', done => {
      main.delete(7).then(() =>
        main.find().then(list => {
          const _list = JSON.parse(localStorage.getItem('snap:overrides:list'));
          // console.log(Object.keys(_list));
          // console.log(_list[7]);
          expect(_list[7]._deleted).to.exist;
          expect(findIndex(list, it => it._id == 7)).to.equal(-1);
          done();
        })
      )
      .catch(err => done(err));
    });
    it('should be also able to delete remote item', done => {
      main.delete(1).then(() =>
        main.find().then(list => {
          expect(findIndex(list, it => it._id == 1)).to.equal(-1);
          done();
        })
      )
      .catch(err => done(err));
    });
  }); 
  describe('constructor', () => {
    let main;
    let list;
    beforeEach(function (done) {
      const list = JSON.parse(localStorage.getItem('snap:overrides:list'));
      list['3'] = {
        _detal: '',
        title: 'Hello',
        preview: 'The third blog Hi there',
        creationTime: 1500097896308,
      };
      storage.set('overrides:list', list);
      main = new Snapmongo('main');
      main.find().then(() => {
        done();
      })
      .catch(err => done(err));
    });
    it('should delete identical local document when the remote has arrived', () => {
      const list = JSON.parse(localStorage.getItem('snap:overrides:list'));
      // console.log(Object.keys(list));
      expect(Object.keys(list).indexOf('3')).to.equal(-1);
    });
  }); 
  describe('find', () => {
    let main;
    before(done => {
      main = new Snapmongo('main');
      main.find().then(() => {
      const keys = Object.keys(localStorage);
      const key = find(keys, key => key.indexOf('/main/index') > 0);
      let str = localStorage.getItem(key);
      str = str.replace(/\{"list([0-9]+)":[0-9]+\}/, `{"list$1":${new Date().getTime()}}`);
      localStorage.setItem(key, str);
        done();
      })
    });
    it('should retrieve list again if timestamp has updated', (done) => {
      main.find().then((res) => {
        const calls = spyOnFetch.__spy.calls;
        const [lastUrl] = calls.slice(-1)[0];
        // console.log(lastUrl);
        expect(lastUrl.startsWith('/main/list1501765968589.json?')).to.be.true;
        done();
      })
      .catch(err => done(err));
    });
  }); 
});
